#!/usr/bin/python
# -*- coding: iso-8859-1 -*-
# A python HTTP Server for Enigform Session Management
# by Arturo 'Buanzo' Busleiman <buanzo@buanzo.com.ar>
# Copyleft 2009 GNU GPLv2

import sys
import gpgme
import hashlib
import datetime
from pysqlite2 import dbapi2 as sqlite
from BaseHTTPServer import BaseHTTPRequestHandler,HTTPServer
from StringIO import StringIO

def formatExceptionInfo():
  """ formatExceptionInfo extracts and returns() the name and value of an exception """
  cla, exc, trbk = sys.exc_info()
  excName = cla.__name__
  try:
    excArgs = exc.__dict__["args"]
  except KeyError:
    excArgs = str(exc)
  return (excName, excArgs)

def msgerror(msg):
  e1,e2 = formatExceptionInfo()
  print msg
  print "Error Details: " + e1 + ":" + e2

def get_session_hash(r):
  sha1 = hashlib.sha1()
  keyid = r.headers['X-Auth-OpenPGP-KeyID']
  sig = r.headers['X-OpenPGP-Sig']
  ua = r.headers['User-Agent']
  sha1.update(keyid)
  sha1.update(sig)
  sha1.update(ua)
  return sha1.hexdigest()

def test_sqlite_session_db():
# http://docs.python.org/library/sqlite3.html
  try:
    connection = sqlite.connect('/tmp/enigform.db')  
    cursor = connection.cursor()
    cursor.execute('CREATE TABLE IF NOT EXISTS sessions (id INTEGER PRIMARY KEY, key TEXT, sh TEXT, created TIMESTAMP)')
    connection.commit()
    connection.close()
  except:
    return False
  return True

def store_session(r,keyid,sh):
  try:
    t = (keyid,sh,datetime.datetime.now())
    connection = sqlite.connect('/tmp/enigform.db')  
    cursor = connection.cursor()
    cursor.execute("DELETE FROM sessions WHERE key = ?",(keyid,))
    cursor.execute("INSERT INTO sessions(key,sh,created) VALUES (?,?,?)",t)
    connection.commit()
    connection.close()
  except:
    return False
  return True

def gpg_encrypt_sign(gpg_data,keyid,want_ascii):
  try:
    ctx = gpgme.Context()
    if want_ascii == 1:
      ctx.armor = True
    else:
      ctx.armor = False
    s_key = ctx.get_key('B2725EC34C2B05D4F0DC25EA6126D64636922767')
    ctx.signers = [s_key]
    key = ctx.get_key(keyid)
    plain = StringIO(gpg_data)
    cipher = StringIO()
    ctx.encrypt_sign([key],gpgme.ENCRYPT_ALWAYS_TRUST,plain,cipher)
    encrypted = cipher.getvalue()
  except:
    msgerror("Begin-Session gpg_encrypt error:")
    return None
  return (encrypted,len(encrypted))

class EnigformPythonServer(BaseHTTPRequestHandler):
  def do_GET(self):
    try:
      client_ip = self.headers['X-Forwarded-For']
    except:
      self.send_error(501,'Server error: %s' % self.path)
      print "No X-Forwarded-For header. You are NOT facing EPS to the Internet, aren't you?"
      return
    try:
      if self.path == '/HTTP_OPENPGP_DISCOVERY': # Enigform Discovery
        print self.path + ' -> Discovery Method'
        self.send_response(200)
        self.send_header('X-OpenPGP-Agent','mod_openpgp with eps/1.0')
        self.send_header('Content-type','text/plain')
        self.end_headers()
        self.wfile.write('HTTP-OpenPGP-Supported')
        return
      elif self.path == '/HTTP_OPENPGP_BEGIN_SESSION': # Secure Session Initiation
        print self.path + ' -> Begin Session Method'
        try:
          xauthopenpgp = self.headers['X-Auth-OpenPGP']
        except:
          print 'No X-Auth-OpenPGP header found. Returning 501/HTTP.'
          self.send_error(501,'Server error: %s' % self.path)
          return
        if xauthopenpgp == 'false':
          print 'Unknown user. X-Auth-OpenPGP is false.'
          self.send_error(501,'Server error: %s' % self.path)
          return
        if xauthopenpgp == 'true':
          # mod_openpgp says we have a signed begin_session request, so handle it
          print 'Begin-Session: Getting Session Hash...'
          sh = get_session_hash(self)
          print 'Begin-Session: SH=%s' % sh
          try:
            keyid = self.headers['X-Auth-OpenPGP-KeyID']
            esh,nbytes = gpg_encrypt_sign(sh,keyid,1)
          except:
            msgerror("Begin-Session Encryption.")
            return
          print 'Begin-Session: Encrypted SH=%s' % esh
          store_session(self,self.headers['X-Auth-OpenPGP-KeyID'],sh)
          print 'Begin-Session: Storage complete'
          self.send_response(200)
          self.send_header('X-OpenPGP-Agent','mod_openpgp with eps/1.0')
          self.send_header('Content-type','text/plain')
          self.send_header('X-Auth-OpenPGP',self.headers['X-Auth-OpenPGP'])
          self.end_headers()
          self.wfile.write(esh)
          return
        print 'Unknown X-Auth-OpenPGP Value!'
        self.send_error(501,'Server error: %s' % self.path)
        return
      elif self.path == '/HTTP_OPENPOP_END_SESSION': # Secure Session Destruction
        print self.path + ' -> End Session Method'
        self.send_response(200)
        self.send_header('Content-type','text/plain')
        self.end_headers()
        self.wfile.write('End Session Method: '+self.path)
        return
    except:
      self.send_error(501,'Unknown error or method called on EPS: %s' % self.path)
      print 'Unknown error/method on EPS: %s' % self.path

  def do_POST(self):
    global rootnode
    try:
      print self.path
      ctype, pdict = cgi.parse_header(self.headers.getheader('content-type'))
      if ctype == 'multipart/form-data':
        query=cgi.parse_multipart(self.rfile, pdict)
        self.send_response(301)
        self.end_headers()
        upfilecontent = query.get('upfile')
        print "filecontent", upfilecontent[0]
        self.wfile.write("<HTML>POST OK.<BR><BR>");
        self.wfile.write(upfilecontent[0]);

    except :
      pass

def main():
  if test_sqlite_session_db() == False:
    print 'Enigform Session Database does not exist. Should create'
    return
  else:
    print 'Enigform Session Database OK.'
  try:
    server = HTTPServer(('127.0.0.1', 2280), EnigformPythonServer)
    print 'EPS has started. Listening on port 2280.'
    server.serve_forever()
  except KeyboardInterrupt:
    print '\n^C received, shutting down server'
    server.socket.close()

if __name__ == '__main__':
  main()
